#!/bin/bash
# (C) Klaus Knopper Jul 2010
# Calls scanpartitions as root and adds entries to /etc/fstab

PATH="/bin:/sbin:/usr/bin:/usr/sbin"
export PATH
umask 022

force=""
forensic=""
verbose=""
remove=""
mkicons=""
user=""
group=""
mpowner="root"
mpgroup="disk"
arg="$1"
while [ -n "$arg" ]; do
 case "$arg" in
  -f*) force="yes" ;;
  -v*) verbose="yes" ;;
  -r*) remove="yes" ;;
  -i*) mkicons="yes" ;;
  -u*) shift; user="$1"; mpowner="$1" ;;
  -g*) shift; group="$1"; mpgroup="$1" ;;
  *) echo "Usage: $0 [-f[orce]] [-v[erbose]] [-r[emove_old]] [-i[cons]] [-u[ser] uid] [-g[roup] gid]" ;;
 esac
 shift
 arg="$1"
done

case "$(cat /proc/cmdline)" in
 *\ nofstab)
  if [ -z "$force" ]; then
   [ -n "$verbose" ] && echo "$0: nofstab bootoption detected, exiting." >&2
   exit 0
  fi
  ;;
 *\ forensic)
  forensic="true"
  ;;
esac

[ ! -r /proc/partitions ] && { echo "$0: /proc not mounted, exiting" >&2; exit 1; }

MYPID="$$"

# Check if it's already running, wait for completion
while 2>/dev/null read PID < /var/run/rebuildfstab.pid; do
 if [ -n "$PID" ] && test -d /proc/"$PID" >/dev/null 2>&1; then
  sleep 2
  continue
 fi
 # Old process gone, but file still there?
 break
done

# Lock.
echo "$MYPID" >/var/run/rebuildfstab.pid

TMP="/tmp/fstab.$$.tmp"
ADDEDBYKNOPPIX="# Added by KNOPPIX"

# Simple shell grep, searches for lines STARTING with string
stringinfile(){
 while read line; do
  case "$line" in $1*) return 0;; esac
 done <"$2"
 return 1
}

removeentries(){
 # Remove comment line $1 and the following line from file $2
 # sed '/^# Added by KNOPPIX/{N;d;}'
 while read line; do
  case "$line" in $1) read line; continue ;; esac
  echo "$line"
 done <"$2"
}

# remove_fstab_entry device
remove_fstab_entry(){
 # Remove comment line $1 and the following line from /etc/fstab
 sed -i -e "/^$ADDEDBYKNOPPIX/{N;/${1//\//\/} /d}" /etc/fstab
}

# Add or replace udev-generated entry to/in /etc/fstab
# add_fstab_entry device [mountpoint]
add_fstab_entry(){
 local device="$1" mountpoint="$2" fstype="$ID_FS_TYPE" options="noauto,users,exec"
 [ -n "$forensic" ] && options="ro"
 [ -n "$fstype" ] || fstype="auto"
 [ -n "$mountpoint" ] || { mountpoint="${device#/dev/}"; mountpoint="/media/${mountpoint}"; }
 case "$fstype" in
  ntfs*|vfat|msdos)           options="${options},umask=000"
   [ "$fstype" = "vfat" ]  && options="${options},shortname=winnt"
   [ "$fstype" = "ntfs" ]  && chgrp fuse "$device"
   [ -n "$user" ]          && options="$options,uid=$user"
   [ -n "$group" ]         && options="$options,gid=$group" ;;
  swap) [ -n "$forensic" ] && return 0 || options="defaults" ;;
 esac
 cat >>/etc/fstab <<.
$ADDEDBYKNOPPIX
$device $mountpoint $fstype $options 0 0
.
}

### MAIN
###
### We now have two operation modes:
### A) do what udev tells us (add/remove partition entries) if DEVNAME and ACTION is set
### B) scan and add entries automatically

if [ -n "$DEVNAME" -a -n "$ACTION" ]; then
 if [ -n "$DEVTYPE" -a "$DEVTYPE" = "disk" ]; then
  # Silently ignore whole disks with partition tables
  if [ -n "$UDISKS_PARTITION_TABLE" ]; then
   rm -f /var/run/rebuildfstab.pid
   exit 0
  fi
 fi
 # Also ignore Raid or LVM members, and extended partitions
 if [ -n "$MD_LEVEL" -o "$ID_FS_TYPE" = "linux_raid_member" -o -n "$DM_UUID" -o "$UDISKS_PARTITION_TYPE" = "0x05" ]; then
  rm -f /var/run/rebuildfstab.pid
  exit 0
 fi
 MOUNTPOINT="${DEVNAME#/dev/}"; MOUNTPOINT="/media/${MOUNTPOINT}"
 case "$ACTION" in
  add|change)
   [ -n "$forensic" ] && blockdev --setro "$DEVNAME" >/dev/null 2>&1
   if [ "$ID_FS_TYPE" = "swap" ]; then
    MOUNTPOINT="none"
   else
    [ -d "$MOUNTPOINT" ] || mkdir -m 755 -p "$MOUNTPOINT"
    chown "$mpowner"."$mpgroup" "$MOUNTPOINT" 2>/dev/null
    compat="/mnt/${MOUNTPOINT##*/}"
    [ -d "$compat" -o -L "$compat" ] || ln -sf "$mountpoint" "$compat" 2>/dev/null
   fi
   # Remove auto-generated entry
   remove_fstab_entry "$DEVNAME"
   # Add new entry (if no manually generated is present)
   grep -q "^$DEVNAME" /etc/fstab || add_fstab_entry "$DEVNAME" "$MOUNTPOINT"
   ;;
  remove)
   if grep -q "^$DEVNAME " /proc/mounts; then
    umount -drf "$DEVNAME" >/dev/null 2>&1
    umount -l "$DEVNAME" >/dev/null 2>&1
   fi
   [ -d "$MOUNTPOINT" ] && rmdir "$MOUNTPOINT" >/dev/null 2>&1
   remove_fstab_entry "$DEVNAME" "$MOUNTPOINT"
   ;;
 esac
else # Non-udev behaviour
 [ -n "$verbose" ] && echo "Scanning for new harddisks/partitions..." >&2
 rm -f "$TMP"
 if [ -n "$remove" ]; then
  removeentries "$ADDEDBYKNOPPIX" /etc/fstab >"$TMP"
 else
  cat /etc/fstab >"$TMP"
 fi
 count=0
 while read device mountpoint fstype relax; do
  stringinfile "$device " "$TMP" || \
  { count="$((count + 1))"
    if [ -n "$mountpoint" -a "$mountpoint" != "none" -a ! -d "$mountpoint" ]; then
     mkdir -p "$mountpoint" 2>/dev/null
     chown "$mpowner"."$mpgroup" "$mountpoint" 2>/dev/null
     chmod 775 "$mountpoint" 2>/dev/null
     compat="/mnt/${mountpoint##*/}"
     [ -d "$compat" -o -L "$compat" ] || ln -sf "$mountpoint" "$compat" 2>/dev/null
    fi
    options="noauto,users,exec"
    case "$fstype" in
     vfat)  options="${options},umask=000,shortname=winnt" ;;
     msdos) options="${options},umask=000" ;;
     swap)  options="defaults" ;;
     ntfs*) options="${options},umask=000"; chgrp fuse "$device";;
    esac
    case "$fstype" in
    ntfs*|vfat|msdos)
    [ -n "$user" ] && options="$options,uid=$user"
    [ -n "$group" ] && options="$options,gid=$group"
    ;;
    esac
    echo "$ADDEDBYKNOPPIX"
    echo "$device $mountpoint $fstype $options 0 0"; }
 done >>"$TMP" <<EOT
$(scanpartitions)
EOT

 # Add CD/DVD-Roms
 cdcount=0
 while read drive name drives; do
  case "$drive$name" in drivename:)
   for d in $drives; do
    mountpoint="/media/$d"
    if [ -n "$mountpoint" -a "$mountpoint" != "none" -a ! -d "$mountpoint" ]; then
     mkdir -p "$mountpoint" 2>/dev/null
     compat="/mnt/${mountpoint##*/}"
     [ -d "$compat" -o -L "$compat" ] || ln -sf "$mountpoint" "$compat" 2>/dev/null
    fi
    stringinfile "/dev/$d " "$TMP" || \
     { cdcount="$((cdcount + 1))"
       echo "$ADDEDBYKNOPPIX"
       echo "/dev/$d $mountpoint auto ro,noauto,users,exec 0 0"; }
   done
   break
   ;;
  esac
 done >>"$TMP" <<EOT
$(cat /proc/sys/dev/cdrom/info 2>/dev/null)
EOT

 [ -n "$verbose" ] && { [ "$count" -gt 0 -o "$cdcount" -gt 0 ] && echo "Adding $count new partitions and $cdcount new cdrom-type devices to /etc/fstab." >&2 || echo "No new partitions found." >&2; }
 mv -f "$TMP" /etc/fstab
fi

 [ -n "$mkicons" ] && type -p mkdesktophdicons  >/dev/null 2>&1 && mkdesktophdicons >/dev/null 2>&1

 rm -f /var/run/rebuildfstab.pid

[ -n "$DISPLAY" ] && sleep 1

exit 0
